@@ -35,6 +35,7 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+#include "evdev.h"
 
 #include <xf86.h>
 #include <xf86Xinput.h>
@@ -42,11 +43,8 @@
 #include <exevents.h>
 
 #include <evdev-properties.h>
-#include "evdev.h"
 
-#ifdef HAVE_PROPERTIES
 static Atom prop_dlock     = 0; /* Drag lock buttons. */
-#endif
 
 void EvdevDragLockLockButton(InputInfoPtr pInfo, unsigned int button);
 
@@ -105,13 +103,12 @@ EvdevDragLockPreInit(InputInfoPtr pInfo)
                     /* We only have a meta button */
                     pEvdev->dragLock.meta = meta_button;
 
-                    xf86Msg(X_CONFIG, "%s: DragLockButtons : "
-                            "%i as meta\n",
-                            pInfo->name, meta_button);
+                    xf86IDrvMsg(pInfo, X_CONFIG, "DragLockButtons : "
+                                "%i as meta\n", meta_button);
                 } else {
-                    xf86Msg(X_ERROR, "%s: DragLockButtons : "
-                            "Incomplete pair specifying button pairs %s\n",
-                            pInfo->name, option_string);
+                    xf86IDrvMsg(pInfo, X_ERROR, "DragLockButtons : "
+                                "Incomplete pair specifying button pairs %s\n",
+                                option_string);
                 }
             } else {
 
@@ -119,23 +116,24 @@ EvdevDragLockPreInit(InputInfoPtr pInfo)
                 if ((meta_button <= EVDEV_MAXBUTTONS) && (meta_button >= 0 ) &&
                     (lock_button <= EVDEV_MAXBUTTONS) && (lock_button >= 0)) {
 
-                    xf86Msg(X_CONFIG, "%s: DragLockButtons : %i -> %i\n",
-                            pInfo->name, meta_button, lock_button);
+                    xf86IDrvMsg(pInfo, X_CONFIG,
+                                "DragLockButtons : %i -> %i\n",
+                                meta_button, lock_button);
 
                     pEvdev->dragLock.lock_pair[meta_button - 1] = lock_button;
                     pairs=TRUE;
                 } else {
                     /* Let the user know something was wrong
                        with this pair of buttons */
-                    xf86Msg(X_CONFIG, "%s: DragLockButtons : "
-                            "Invalid button pair %i -> %i\n",
-                            pInfo->name, meta_button, lock_button);
+                    xf86IDrvMsg(pInfo, X_CONFIG,"DragLockButtons : "
+                                "Invalid button pair %i -> %i\n",
+                                meta_button, lock_button);
                 }
             }
         } else {
-            xf86Msg(X_ERROR, "%s: Found DragLockButtons "
-                    "with  invalid lock button string : '%s'\n",
-                    pInfo->name, option_string);
+            xf86IDrvMsg(pInfo, X_ERROR, "Found DragLockButtons "
+                        "with invalid lock button string : '%s'\n",
+                        option_string);
 
             /* This should be the case anyhow, just make sure */
             next_num = NULL;
@@ -145,6 +143,8 @@ EvdevDragLockPreInit(InputInfoPtr pInfo)
         if (next_num != NULL && *next_num == '\0')
             next_num = NULL;
     }
+
+    free(option_string);
 }
 
 /* Updates DragLock button state and fires button event messges */
@@ -209,7 +209,6 @@ EvdevDragLockFilterEvent(InputInfoPtr pInfo, unsigned int button, int value)
     return FALSE;
 }
 
-#ifdef HAVE_PROPERTIES
 /**
  * Set the drag lock property.
  * If only one value is supplied, then this is used as the meta button.
@@ -256,7 +255,7 @@ EvdevDragLockSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
                 pEvdev->dragLock.meta = meta;
                 memset(pEvdev->dragLock.lock_pair, 0, sizeof(pEvdev->dragLock.lock_pair));
             }
-        } else
+        } else if ((val->size % 2) == 0)
         {
             CARD8* vals = (CARD8*)val->data;
 
@@ -269,10 +268,11 @@ EvdevDragLockSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
                 pEvdev->dragLock.meta = 0;
                 memset(pEvdev->dragLock.lock_pair, 0, sizeof(pEvdev->dragLock.lock_pair));
 
-                for (i = 0; i < val->size && i < EVDEV_MAXBUTTONS; i++)
-                    pEvdev->dragLock.lock_pair[i] = vals[i];
+                for (i = 0; i < val->size && i < EVDEV_MAXBUTTONS; i += 2)
+                    pEvdev->dragLock.lock_pair[vals[i] - 1] = vals[i + 1];
             }
-        }
+        } else
+            return BadMatch;
     }
 
     return Success;
@@ -316,5 +316,3 @@ EvdevDragLockInitProperty(DeviceIntPtr dev)
 
     XIRegisterPropertyHandler(dev, EvdevDragLockSetProperty, NULL, NULL);
 }
-
-#endif
